#!/usr/bin/env node
// Copyright IBM Corp. 2016. All Rights Reserved.
// Node module: apiconnect
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

var Command = require('commander').Command;
var Promise = require('bluebird');
var _ = require('lodash');
var d = require('debug')('apiconnect:lib:cli-builder');
var format = require('util').format;
var fs = require('fs');
var config = require('apiconnect-cli-config').loadConfig();
var netrc = require('apiconnect-cli-util').Netrc;
var ServerError = require('apiconnect-cli-util').ServerError;
var logger = require('apiconnect-cli-logger');
var path = require('path');
var util = require('util');

module.exports = function(plugins) {
  return buildCli(buildCommandMap(plugins));
};

var DEFAULT_TYPE = -1;
function buildCommandMap(plugins) {
  var commandList = {};

  _.each(plugins, function(plugin) {
    _.each(plugin.topics, function(topic) {
      _.each(topic.commands, function(command) {
        command.topic = topic.name;
        command.description = topic.description;
        command.parentHelpInfo = topic.helpInfo;
        command.isBuiltin = plugin.isBuiltin;
        var verb = command.command.split(' ')[0];
        var fullName = topic.name + ':' + verb;
        var type = command.type || DEFAULT_TYPE;
        if (type !== DEFAULT_TYPE) {
          command.typeHelpInfo = command.helpInfo;
        }
        commandList[fullName] = commandList[fullName] || {};
        commandList[fullName][type] = command;
      });
    });
  });

  return commandList;
}

Command.prototype.version = function(str, flags) {
  if (arguments.length === 0) {
    return this._version;
  }
  this._version = str;
  flags = flags || '-V, --version';
  this.option(flags, 'output the version number');
  this.on('version', function() {
    process.stdout.write(str + '\n');
  });
  return this;
};

Command.prototype.helpInformation = function() {
  var maxWidth = 0;

  var commands = _.filter(this.commands, function(command) {
    return command.topic && command.topic.length > 0 && !command.isBuiltin;
  }).map(function(command) {
    if (command.topic.length > maxWidth) {
      maxWidth = command.topic.length;
    }
    var description = _.isFunction(command.description) ?
      command.description() : command.description;
    return {
      name: command.topic,
      description: description || '',
    };
  });

  commands = _.uniqWith(commands, _.isEqual);

  var apicHelp = fs.readFileSync(path.resolve(__dirname, path.join('..', 'intl', 'en', 'apic_help.txt')), 'utf8');
  var help = [
    '',
    apicHelp,
    '',
  ];

  if (commands.length > 0) {
    help = help.concat([
      'Additional topics',
      '',
    ]);
    help = help.concat(commands.map(function(command) {
      return '  ' + _.padEnd(command.name, maxWidth + 4) + command.description;
    }));
    help = help.concat([ '' ]);
  }
  return help.join('\n');
};

function buildCli(commandList) {
  var cli = new Command();
  var cl1 = _.flatten(_.map(commandList, _.values));
  var cl2 = _.flatten(_.map(cl1, 'aliases'));
  var allAliases = _.uniqWith(_.remove(cl2, function(n) {
    return _.isString(n);
  }), _.isEqual);

  cli.allAliases = allAliases;

  _.each(commandList, function(types, name) {
    if (Object.keys(types).length === 1 && types.hasOwnProperty(DEFAULT_TYPE)) {
      // simple command
      buildCommand(cli, name, types[DEFAULT_TYPE], false, allAliases);
    } else {
      var c = cli.command(name);

      // build sub-command by types
      var subcommands = {};
      var aliases = [];
      var parentHelpInfo = null;
      _.each(types, function(commandInfo, type) {
        var subcli = new Command();
        buildCommand(subcli, name, commandInfo, true, allAliases);
        aliases = aliases.concat(commandInfo.aliases);
        subcommands[type] = subcli;
        if (commandInfo.typeHelpInfo) {
          subcommands[type].helpInfo = commandInfo.typeHelpInfo;
        }
        if (commandInfo.parentHelpInfo) {
          parentHelpInfo = commandInfo.parentHelpInfo;
        }
        if (commandInfo.defaultType) {
          if (commandInfo.typeHelpInfo) {
            parentHelpInfo = commandInfo.typeHelpInfo;
          }
          c.defaultType = subcli;
        }
      });

      c.option('-e, --type <type>', '');
      c.outputHelp = function() {
        if (this.type && subcommands[this.type].helpInfo) {
          printHelp(subcommands[this.type]);
        } else if (parentHelpInfo) {
          printHelp(parentHelpInfo);
        }
      };
      _.each(aliases, function(alias) {
        c.alias(alias);
      });
      c.action(function() {
        var o = arguments[arguments.length - 1];
        if (o.type && !subcommands[o.type]) {
          throw Error(format(
            '`%s` is not a valid type for command %s. Possible type values are [%s]',
            o.type, name, _.keys(subcommands)
          ));
        }
        var type = o.type || DEFAULT_TYPE;
        var sc = subcommands[type] || c.defaultType;
        if (!sc) {
          if (parentHelpInfo) {
            printHelp(parentHelpInfo);
          } else {
            throw Error(format('The command %s expect a type. Possible type values are [%s]',
                        name, _.keys(subcommands)));
          }
        } else {
          logger.debug(format('invoking subcommand with type: %s', type));
          sc.parse(process.argv);
        }
      });
      c.allowUnknownOrMissingArgs = true;
      c.allowUnknownOption();
    }
  });

  return cli;
}

function buildCommand(cli, name, commandInfo, addTypeOption, allAliases) {
  d('buildCommand %s %j', name, commandInfo);
  commandInfo.allowExtraArguments = !!commandInfo.allowExtraArguments;
  commandInfo.fullName = name;
  var c = cli.command(commandInfo.topic + ':' + commandInfo.command);
  c.topic = commandInfo.topic;
  c.description = commandInfo.description;
  _.each(commandInfo.aliases, function(alias) {
    c.alias(alias);
  });
  _.each(commandInfo.options, function(option) {
    c.option(option, '');
  });
  if (addTypeOption) {
    c.option('-e, --type <type>', '');
  }
  c.action(actionWrapper(commandInfo));
  c.outputHelp = function() {
    printHelp(commandInfo);
  };
  c.isBuiltin = commandInfo.isBuiltin;
  if (!commandInfo.allowUnknownOrMissingArgs) {
    d('unknown options or args will not be allowed');
    c.optionMissingArgument = function(option, flag) {
      optionMissingArgument(option, flag);
    };
    c.unknownOption = function(flag) {
      unknownOption(flag);
    };
    c.missingArgument = function(name) {
      missingArgument(name);
    };
  } else {
    c.allowUnknownOption();
  }
  // TODO Remove after plugins are updated to use
  // `apiconnect-cli-logger` directly
  c.logger = logger;

  var allCommands = _.map(cli.commands, '._name');
  if (allAliases.length) {
    allCommands = allCommands.concat(allAliases);
  }
  // create command for plugin only if there are no aliases of the same name
  if (_.indexOf(allCommands, commandInfo.topic) < 0) {
    var cp = cli.command(commandInfo.topic);
    cp.isBuiltin = commandInfo.isBuiltin;
    cp.outputHelp = function() {
      printHelp(commandInfo.parentHelpInfo);
    };
    cp.action(cp.outputHelp);
  };
}

function missingArgument(name) {
  var err = new Error(format('missing required argument `%s`', name));
  err.invalidCommand = true;
  throw err;
}

function unknownOption(flag) {
  var err = new Error(format('unknown option `%s`', flag));
  err.invalidCommand = true;
  throw err;
}

function printHelp(helpObject) {
  var helpText = '';
  if (helpObject.helpInfo) {
    helpText = _.isFunction(helpObject.helpInfo) ? helpObject.helpInfo() : helpObject.helpInfo;
  } else {
    helpText = _.isFunction(helpObject) ? helpObject() : helpObject;
  }
  logger.writeln(helpText);
  logger.exit(0);
}

function optionMissingArgument(option, flag) {
  var err;
  if (flag) {
    err = new Error(format('option `%s` argument missing, got `%s`', option.flags, flag));
    err.invalidCommand = true;
    throw err;
  } else {
    err = new Error(format('option `%s` argument missing', option.flags));
    err.invalidCommand = true;
    throw err;
  }
}

function actionWrapper(cmd) {
  return function() {
    logger.debug('invoking %s', cmd.fullName);
    var args = Array.apply(null, arguments);
    // If we were passed more arguments than necessary, strip out extras
    if (!cmd.allowExtraArguments && args.length !== cmd.action.length) {
      logger.debug('num args (%s) does not match expected (%s)', args.length, cmd.action.length);
      var opts = arguments[arguments.length - 1];
      // Trim the args to the size expected by the command
      args = args.slice(0, cmd.action.length);
      // The last argument needs to be the opts object. Put that one back in
      args[args.length - 1] = opts;
    }

    // If loginRequired is set in cmd, ensure we are logged in.
    if (cmd.loginRequired) {
      checkLogin(cmd, args[args.length - 1]);
    }

    Promise.resolve(cmd.action.apply(this, args))
      .then(function(res) {
        logger.debug('%s completed succesfully with results %j', cmd.fullName, res);
        logger.exit(0);
      })
      .catch(function(err) {
        if (err instanceof ServerError) {
          logger.error(err.toString());
        } else {
          logger.error(err);
        }

        logger.debug('%s completed with error', cmd.fullName);
        var msg = err && err.exitCode ? err.exitCode : 1;
        logger.exit(msg);
      });
  };
}

function checkLogin(cmd, options) {
  var hostname;
  var hostConfig;
  var type;
  switch (cmd.loginRequired) {
    case 'app':
      hostConfig = config.getOne('app');
      type = 'app';
      break;
    case 'catalog':
    case true:
      hostConfig = config.getOne('catalog');
      type = 'catalog';
      break;
    default:
      return;
  }

  if (hostConfig && hostConfig.hostname) {
    hostname = hostConfig.hostname;
  }
  hostname = options.server || hostname;

  if (!hostname) {
    throw new Error(
      format('--server is required or needs to be set via `apic config:set %s`', type)
    );
  }

  if (netrc.lookup(hostname)) {
    logger.debug('Successful lookup of %s', hostname);
    return;
  } else {
    throw new Error(util.format("Use 'apic login%s' to log into %s", type === 'app' ? ' --type app' : '', hostname));
  }
}
